#!/usr/bin/env python3

'''
Methods for normalization of timeseries data.
'''

from pandas import DataFrame, Series

def standard(df: DataFrame) -> DataFrame:
  '''
    Returns standard normalization of the TimeSeries data.
    @param df: pd.DataFrame
  '''
  if isinstance(df, DataFrame):
    for col in list(df.columns):
      df[col] = (df[col] - df[col].min()) / (df[col].max() - df[col].min())
  elif isinstance(df, Series):
    df = df.apply(lambda x: ((x - df.min()) / df.max() - df.min()))
  else:
    ... # TODO error wrong type
  return df
  
def zscore(df: DataFrame) -> DataFrame:
  '''
    Returns zscore normalization of the TimeSeries data.
    @param df: pd.DataFramee
  '''
  if isinstance(df, DataFrame):
    for col in list(df.columns):
      df[col] = (df[col] - df[col].mean()) / df[col].std(ddof=0)
  elif isinstance(df, Series):
    df = df.apply(lambda x: ((x - df.mean()) / df.std(ddof=0)))
  else:
    ... # TODO error wrong type
  return df


#!/usr/bin/env python3

# Source: https://skemman.is/bitstream/1946/15343/3/SS_MSthesis.pdf

# Largest-Triangle-One-Bucket
'''
Algorithm 4.1 Largest-Triangle-One-Bucket
Require: data . The original data
Require: threshold . Number of data points to be returned
1: Rank every point in regard to its effective area
2: Split the data into the same number of buckets as the threshold
3: for each bucket do
4: Select the point with the highest rank within the bucket
5: end for
6: Finally make sure that the first and last data points in the original data are also
the first and last data points in the downsampled data.
'''
def ltob(raw_timeseries_data, threshold=False):
  if treshold:
    ...

  else:
    ...

  return _timeseries_data

# Largest-Triangle-Three-Buckets
'''
Algorithm 4.2 Largest-Triangle-Three-Buckets
Require: data . The original data
Require: threshold . Number of data points to be returned
1: Split the data into equal number of buckets as the threshold but have the first
bucket only containing the first data point and the last bucket containing only
the last data point
2: Select the point in the first bucket
3: for each bucket except the first and last do
4: Rank every point in the bucket by calculating the area of a triangle it forms
with the selected point in the last bucket and the average point in the next
bucket
5: Select the point with the highest rank within the bucket
6: end for
7: Select the point in the last bucket . There is only one
'''
def lttb(raw_timeseries_data, threshold=False):
  if treshold:
    ...
    
  else:
    ...

  return _timeseries_data

# Largest-Triangle-Dynamic
'''
Algorithm 4.3 Largest-Triangle-Dynamic
Require: data . The original data
Require: threshold . Number of data points to be returned
1: Split the data into equal number of buckets as the threshold but have the first
bucket only containing the first data point and the last bucket containing only
the last data point . First and last buckets are then excluded in the bucket
resizing
2: Calculate the SSE for the buckets accordingly . With one point in adjacent
buckets overlapping
3: while halting condition is not met do . For example, using formula 4.2
4: Find the bucket F with the highest SSE
5: Find the pair of adjacent buckets A and B with the lowest SSE sum . The
pair should not contain F
6: Split bucket F into roughly two equal buckets . If bucket F contains an odd
number of points then one bucket will contain one more point than the other
7: Merge the buckets A and B
8: Calculate the SSE of the newly split up and merged buckets
9: end while.
10: Use the Largest-Triangle-Three-Buckets algorithm on the resulting bucket configuration to select one point per buckets
'''
def ltd(raw_timeseries_data, threshold=False):
  if treshold:
    ...
    
  else:
    ...

  return _timeseries_data


#!/usr/bin/env python3

'''
Methods for univariate descriptive timeseries analysis.
'''

import pandas as pd
from pandas import Series, DataFrame
import math

def simple_descriptive_analysis(series: Series) -> dict:
  '''
    Returns dictionary with TimeSeries specific features.
    This method generates a simple descriptive analysis of the timeseris data.
    TODO ...
    @param series: pd.Series
     
  '''
  result = {
    'min': series.min(),
    'mean': series.mean(),
    'median': series.median(),
    'max': series.max(),
    'std': series.std(),
    'var': series.var(),
    'largest': series.nlargest(),
    'smallest': series.nsmallest()
  }
  return result

def rolling_mean(series: Series, window=5) -> Series:
  '''
    Returns a Series with the rolling mean.
    @param series: pd.Series
  '''
  return series.rolling(window, min_periods=1).mean()

def rolling_std(series: Series, window=5) -> Series:
  '''
    Returns a Series with the rolling std.
    @param series: pd.Series
  '''
  return series.rolling(window, min_periods=1).std()

def confidence_interval(series: Series) -> DataFrame:
  '''
    Returns a DataFrame with three columns, the upper and lower bound and the rolling mean for the confidence interval of the input timeseries.
    @param series: pd.Series
  '''
  result_df = pd.DataFrame()
  result_df['mean'] = rolling_mean(series)
  buf_std = rolling_std(series)
  ci95_hi = []
  ci95_lo = []
  for i in range(0,len(result_df.index)):
    ci95_hi.append(result_df['mean'][i] + 1.95*buf_std[i]/math.sqrt(5))
    ci95_lo.append(result_df['mean'][i] - 1.95*buf_std[i]/math.sqrt(5))
  
  result_df['ci_hi'] = ci95_hi
  result_df['ci_lo'] = ci95_lo
  return result_df

def value_distribution(series: Series, bins=-1) -> Series:
  '''
    Returns a DataFrame with the value distribution of the input timeseries (rounded).
    @param series: pd.Series
  '''
  if bins==-1:
    return series.round().value_counts().sort_index()
  else:
    return series.round().value_counts(bins=bins).sort_index()


#!/usr/bin/env python3

'''
Methods for visualizations of timeseries data based on the seaborn package.
'''

from pandas import Series
import seaborn as sns
import matplotlib.pyplot as plt

def sns_line_plot_series(series: Series, show=True) -> None:
  sns.lineplot(data=series)
  if show:
    plt.show()
  # TODO generate and return graphic
  return

def sns_line_plot_ci(series: Series, show=True) -> None:
  # TODO Area plot between ci_hi and ci_lo
  sns.lineplot(data=series)
  if show:
    plt.show()
  # TODO generate and return graphic
  return


#!/usr/bin/env python3

'''
This is the hctsa-core module, using Flask as a minimal backend webserver.
'''

# TODO Flask Server
# rest API:
# - potentielle nachfolgermethoden basierend auf spezifischer methode, array mit potentiellen methoden
# - add_method, return true, false
# - pipeline löschen/neu 'starten', return true, false
# - pipeline letztes element löschen, return true, false
# - pipeline run, return results

from hctsa.pipeline.pipeline import Pipeline
from flask import Flask, request, jsonify
from flask_cors import CORS

app = Flask(__name__)
CORS(app)

pipeline = Pipeline()

@app.route("/")
def hello_world():
    return "<p>Hello, World!</p>"

@app.route("/pipeline/rules", methods=['GET'])
def pipeline_rules():
    return jsonify(pipeline.method_rules)

@app.route("/pipeline/data", methods=['POST'])
def pipeline_data():
  filename = request.get_json()['filename']
  # TODO get data and store into pipeline
  return pipeline.load_data_csv(filename=filename)

@app.route("/pipeline/add", methods=['POST'])
def pipeline_add():
  add_function = request.get_json()['method']
  # TODO add parameters... e.g. visualize as graphic and return the graphic...
  if pipeline.add_method(add_function):
    return add_function + ' added'
  return add_function + ' not added - error'

@app.route("/pipeline/delete", methods=['GET', 'POST'])
def pipeline_delete():
  delete_position = request.get_json()['position']
  pipeline.del_method(delete_position)
  return 'method on position ' + str(delete_position) + ' deleted'

@app.route("/pipeline/run", methods=['GET', 'POST'])
def pipeline_run():
  _result = pipeline.run()
  return jsonify({'result': _result.to_json()})

@app.route("/pipeline/reset", methods=['GET', 'POST'])
def pipeline_reset():
  pipeline.reset_pipeline()
  return 'pipeline resetted'

if __name__ == '__main__':
    app.run(host='0.0.0.0', port=5000)



#!/usr/bin/env python3

'''
Pipeline Class, for building a user defined pipeline for timeseries manipulation and analysis.
'''


import pkg_resources
import yaml
import pandas as pd

import hctsa.pipeline.methods as methods

class Pipeline():
  # Main Pipeline Array
  main_pipeline = []
  merge_pipeline = []

  # Main Data Variable
  core_data = []

  # List of additional Data
  add_data = []

  # DICTIONARIE for specific pipeline rules, loaded from method_rules.yml file.
  method_rules = []

  method_rules_source = pkg_resources.resource_filename(__name__, './method_rules.yml')
  
  def __init__(self, series: pd.Series = None) -> None:
    '''
      Initialization of a new Pipeline Object.
      @param series: pd.Series
    '''
    if series is not None:
      self.core_data = series
    # TODO add try/catch block
    #with open('./pipeline/method_rules.yml', 'r') as file:
    with open(self.method_rules_source, 'r') as file:
      self.method_rules = yaml.safe_load(file)
    return

  def load_data_csv(self, filename: str) -> None:
    '''
      Loads pd.Series data as core_data from filename.
      @param filename: str
    '''
    if filename == 'test':
      self.core_data = pd.read_csv('../test_data/testdata.csv', squeeze=True)
    else:
      self.core_data = pd.read_csv(filename, squeeze=True)
    return filename + ' loaded'

  def add_method(self, add_function: str, position: int = -1) -> bool:
    '''
      Adds a new method into the pipeline
      1) Check actual pipeline
      2) Check if method can be added at defined position or at the end (-1), regarding to the pipeline-rules (e.g. visualisation is the last layer/method)
        -> check if method is in method_rules.yml for specific "nachfolger function"
      3) if possible, add method to method-pipeline-array
      Return boolean - success or error msg
      @param add_function: str
      @param position: int
    '''
    if(len(self.main_pipeline)) > 0:
      last_func = self.main_pipeline[-1]
      if(add_function in self.method_rules['methods'][last_func]):
        self.main_pipeline.append(add_function)
      else:
        print('error: function cannot be added: ', add_function)
        return False
    else:
      if(add_function in self.method_rules['start']):
        self.main_pipeline.append(add_function)
      else:
        print('error: function cannot be added: ', add_function)
        return False
    return True

  def del_method(self, position = -1) -> bool:
    self.main_pipeline.pop(position)
    return True

  def reset_pipeline(self) -> None:
    '''
      Resetting the main_pipeline and core_data to an empty array [].
    '''
    self.main_pipeline = []
    self.core_data = []
    return True

  def add_series(self, series: pd.Series, merge_method: str) -> bool:
    '''
      Add an additional timeseries into the pipeline with a specific method for merging two timeseries data (e.g. calculate the correlation)
    '''
    if self.add_method(merge_method):
      self.add_data.append(series)
    return True

  def run(self) -> pd.Series:
    '''
      Returns the result of the Pipeline.
      Method for running the pipeline. Every method that was added into the pipeline will get executed in a sequential way.
      TODO if _data is empty -> error
    '''
    _data = self.core_data
    print('PIPELINE: ', self.main_pipeline)
    for method in self.main_pipeline:
      print('DOING: ', method)
      if method in self.method_rules['merge_methods']:
        _data = getattr(methods, method)(_data, self.add_data[0])
        del self.add_data[0]
      else:
        _data = getattr(methods, method)(_data)
      print(_data)
    return _data


#!/usr/bin/env python3

'''
Methods to transform univariate timeseries data, a restructuring of the data (e.g., calculate the differences between each point).
'''

from pandas import Series

def point_absdiff_transformation(series: Series) -> Series:
  '''
    Returns calculate the absolute differences between the points as a new transformation of the time series.
    First index of the original time series is lost, as the new one has one value less.
    @param series: pd.Series
  '''
  _values = []
  for i in range(1, len(series)):
    _values.append(abs(series[i] - series[i-1]))
  return Series(data=_values, index=series.index[1:])

def point_slope_transformation(series: Series) -> Series:
  '''
    Returns calculate the slope between the points as a new transformation of the time series.
    First index of the original time series is lost, as the new one has one value less.
    @param series: pd.Series
  '''
  _values = []
  print(series)
  for i in range(1, len(series)):
    _values.append(series[i] - series[i-1])
  return Series(data=_values, index=series.index[0:-1])




#!/usr/bin/env python3

'''
Methods to prepare multivariat timeseries data for multivariat methods/combinations.
'''

# TODO
# Vorbereitung der Zeitreihen zur weiteren Verarbeitung (auf gleiche länge schneiden z.B.)
# Kombinationen von Zeitreihen ('ueberlagerung')

from pandas import Series

def merge_sum(series: Series, add_series: Series) -> Series:
  '''
    Returns a merge Series, representing the sum of the values per index of both Series (series + add_series).
    @param series: pd.Series
  '''
  return series + add_series

def merge_subtract(series: Series, add_series: Series) -> Series:
  '''
    Returns a merge Series, representing the subtract of the values per index of both Series (series - add_series).
    @param series: pd.Series
  '''
  return series - add_series

def merge_multiply(series: Series, add_series: Series) -> Series:
  '''
    Returns a merge Series, representing the product of the values per index of both Series (series * add_series).
    @param series: pd.Series
  '''
  return series * add_series

def merge_divide(series: Series, add_series: Series) -> Series:
  '''
    Returns a merge Series, representing the division of the values per index of both Series (series / add_series).
    @param series: pd.Series
  '''
  return series / add_series

def merge_mean(series: Series, add_series: Series) -> Series:
  '''
    Returns a merge Series, representing the mean of the values per index of both Series((series + add_series) / 2).
    @param series: pd.Series
  '''
  return (series + add_series) / 2


#!/usr/bin/env python3

'''
Method mapping file, needed for the string to method mapping functionality.
'''

# Imports
from ..preparation.normalization import *
from ..preparation.transformation import *
from ..preparation.multivariat import *
from ..descriptive.univariate import *
from ..descriptive.multivariat import *
from ..visualization.seaborn import *

# Preparations.Normalizations
def prep_norm_zscore(df: DataFrame) -> DataFrame:
  return zscore(df)

def prep_norm_standard(df: DataFrame) -> DataFrame:
  return standard(df)

# Preparation.Multivariat
def prep_merge_sum(series: Series) -> Series:
  return merge_sum(series)

def prep_merge_subtract(series: Series) -> Series:
  return merge_subtract(series)

def prep_merge_multiply(series: Series) -> Series:
  return merge_multiply(series)

def prep_merge_divide(series: Series) -> Series:
  return merge_divide(series)

def prep_merge_mean(series: Series) -> Series:
  return merge_mean(series)

# Preparations.Transformation
def prep_transform_point_absdiff_transformation(series: Series) -> Series:
  return point_absdiff_transformation(series)

def prep_transform_point_slope_transformation(series: Series) -> Series:
  return point_slope_transformation(series)

# Descriptive.Univariate
def descr_uni_rolling_mean(series: Series) -> Series:
  return rolling_mean(series)

def descr_uni_rolling_std(series: Series) -> Series:
  return rolling_std(series)

def descr_uni_confidence_interval(series: Series) -> DataFrame:
  return confidence_interval(series)

def descr_uni_value_distribution(series: Series) -> Series:
  return value_distribution(series)

def descr_uni_simple_descriptive_analysis(series: Series) -> dict:
  return simple_descriptive_analysis(series)

# Descriptive.Multivariat
def descr_corr(series: Series, add_series: Series) -> int:
  return corr(series, add_series)

# Visualizations.Seaborn
def viz_sns_line_plot_series(series: Series) -> None:
  # TODO return graphic.. 
  return sns_line_plot_series(series)

def viz_sns_line_plot_ci(series: Series) -> None:
  # TODO return graphic.. 
  return sns_line_plot_ci(series)


#!/usr/bin/env python3

'''
Methods for multivariate descriptive timeseries analysis.
'''

# TODO
# correlation
# Series.corr(other, method='pearson', min_periods=None)
# covarianz

from pandas import Series

def corr(series: Series, add_series: Series, method='pearson') -> int:
  '''
    Returns the correlation between both timeseries, based on method (pearson, kendall, spearman).
    @param series: pd.Series
    @param add_series: pd.Series
    @param method: str
  '''
  return series.corr(add_series, method=method)


#from distutils.core import setup
from setuptools import find_packages, setup

setup(
    name = 'hctsa',
    packages = find_packages(),
    version = '0.1.0',  # Ideally should be same as your GitHub release tag version
    description = 'Human-Centered TimeSeries Analysis Package',
    author = 'Andre Sekulla',
    author_email = 'andreseku@gmail.com',
    url = 'https://github.com/AndreSeku/humancenteredTSA',
    download_url = 'https://github.com/AndreSeku/humancenteredTSA/archive/refs/tags/0.1.0.zip',
    keywords = ['timeseries', 'analysis', 'hcml'],
    classifiers = [],
)